//package agents;

import java.util.ArrayList;
//import java.util.*;
import java.util.HashMap;
import java.util.Random;


import negotiator.Agent;
import negotiator.Bid;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.Offer;
import negotiator.issue.Issue;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.IssueInteger;
import negotiator.issue.IssueReal;
import negotiator.issue.Value;
import negotiator.issue.ValueInteger;
import negotiator.issue.ValueReal;


/**
 * @author W.Pasman
 * Some improvements over the standard SimpleAgent.
 */
public class AgentI extends Agent
{
	private Action actionOfPartner=null;
	private static double MINIMUM_BID_UTILITY = 0.5;
	Bid maxbid = null;
	double maxutil = 0, minutil = 0;
	double sumutil = 0;
	double avrutil = 0;
	int count = 0;
	double etotal = 0, etest = 0;
	double firstutil, pborder = 0;
	double discountFactor=1.0;
    ArrayList<Double> ptnutil = new ArrayList<Double>();
		
	/**
	 * init is called when a next session starts with the same opponent.
	 */
	public void init()
	{
		if(utilitySpace.getReservationValue() != null)
		{
			MINIMUM_BID_UTILITY = utilitySpace.getReservationValue();
		}
		
		try 
	    {
		this.discountFactor=utilitySpace.getDiscountFactor();
		if (discountFactor==0.0){
			discountFactor=1;
		}

  }
  catch(Exception ex) {
		//System.out.println("error when seeting DF");
	
  }
	
		
	
	}

	public static String getVersion() { return "1.2"; }

	public void ReceiveMessage(Action opponentAction) 
	{
		actionOfPartner = opponentAction;
	}

	
	public Action chooseAction()
	{
		Action action = null;
        double RTime=timeline.getTime();
		try 
		{ 
			if(actionOfPartner==null) action = chooseRandomBidAction();
			if(actionOfPartner instanceof Offer)
			{
				Bid partnerBid = ((Offer)actionOfPartner).getBid();
				double offeredutil = getUtility(partnerBid)/Math.pow(discountFactor,RTime);
				// get current time
				double time = timeline.getTime();
				double e = 0,ee = 0;
				
				// get my happy utility & bid
				if(maxbid == null){
					maxbid = partnerBid;
					maxutil = offeredutil;
				}
				if(maxutil < offeredutil){
					maxutil = offeredutil;
					maxbid = partnerBid;
				}
				
				// get my sad utility
				if(minutil > offeredutil){
					minutil = offeredutil;
				}
				
				//get utility average
				sumutil += offeredutil;
				if(count == 0) {
					firstutil = offeredutil;
					pborder = (firstutil - 1) * time + 1;
					}
				
				count++;
				avrutil = sumutil/count;
				//System.out.println("max : "+maxutil);
				//System.out.println("avr : "+avrutil);
		
				ptnutil.add(offeredutil);
				//System.out.println(ptnutil);
				
				// get sample variance 
			    if(time > 0.1){
			    for (int i=0; i<ptnutil.size(); i++) {  
			        e = (double)ptnutil.get(i);
			    	ee = avrutil - e;
			    	ee *= ee;
			    	etotal += ee;
			    }
			    etotal = etotal / count;
				//System.out.println("etotal : "+etotal);
				//System.out.println("etest : "+etest);
			    }     
			    if(time >= 0.10 && time < 0.11) etest = etotal;
			    
			    /*Accept judge from P*/
			    if(offeredutil >= maxutil){
				double P=Paccept(offeredutil,time);
				if(pborder < offeredutil){
					if (P>Math.random()) 
						action = new Accept(getAgentID());
					else action = chooseRandomBidAction();               
					}
				else action = chooseRandomBidAction();               
				}			
			    else action = chooseRandomBidAction(); 
			}
			//sleep(0.005); // just for fun
		} catch (Exception e) { 
			//System.out.println("Exception in ChooseAction:"+e.getMessage());
			action=new Accept(getAgentID()); // best guess if things go wrong. 
		}
		return action;
	}
	

	
	private Action chooseRandomBidAction() 
	{
		Bid nextBid=null ;
		try { nextBid = newBid(); } 
		catch (Exception e) { //System.out.println("Problem with received bid:"+e.getMessage()+". cancelling bidding"); 
			}
		
		if (nextBid == null) return (new Accept(getAgentID()));                
		return (new Offer(getAgentID(), nextBid));
	}
	
	/**
	 * @return a random bid with high enough utility value.
	 * @throws Exception if we can't compute the utility (eg no evaluators have been set)
	 * or when other evaluators than a DiscreteEvaluator are present in the util space.
	 */
	
	
	
	private Bid getRandomBid() throws Exception
	{
		HashMap<Integer, Value> values = new HashMap<Integer, Value>(); // pairs <issuenumber,chosen value string>
		ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
		Random randomnr= new Random();

		// create a random bid with utility>MINIMUM_BID_UTIL.
		// note that this may never succeed if you set MINIMUM too high!!!
		// in that case we will search for a bid till the time is up (2 minutes)
		// but this is just a simple agent.
		Bid bid=null;
		double currentUtil=0.0;
		do 
		{
			for(Issue lIssue:issues) 
			{
				switch(lIssue.getType()) {
				case DISCRETE:
					IssueDiscrete lIssueDiscrete = (IssueDiscrete)lIssue;
					int optionIndex=randomnr.nextInt(lIssueDiscrete.getNumberOfValues());
					values.put(lIssue.getNumber(), lIssueDiscrete.getValue(optionIndex));
					break;
				case REAL:
					IssueReal lIssueReal = (IssueReal)lIssue;
					int optionInd = randomnr.nextInt(lIssueReal.getNumberOfDiscretizationSteps()-1);
					values.put(lIssueReal.getNumber(), new ValueReal(lIssueReal.getLowerBound() + (lIssueReal.getUpperBound()-lIssueReal.getLowerBound())*(double)(optionInd)/(double)(lIssueReal.getNumberOfDiscretizationSteps())));
					break;
				case INTEGER:
					IssueInteger lIssueInteger = (IssueInteger)lIssue;
					int optionIndex2 = lIssueInteger.getLowerBound() + randomnr.nextInt(lIssueInteger.getUpperBound()-lIssueInteger.getLowerBound());
					values.put(lIssueInteger.getNumber(), new ValueInteger(optionIndex2));
					break;
				default: throw new Exception("issue type "+lIssue.getType()+" not supported by SimpleAgent2");
				}
			}
			bid=new Bid(utilitySpace.getDomain(),values);
			currentUtil=getUtility(bid)/Math.pow(discountFactor,timeline.getTime());
		} while (currentUtil < MINIMUM_BID_UTILITY);

		return bid;
	}
	
	
	private Bid newBid() throws Exception{
		double t = timeline.getTime();
		if(t < 0.99){
		Bid bst = null;
		Bid bid = getRandomBid();
		double scr = getUtility(bid)/Math.pow(discountFactor, timeline.getTime());
		double siki = 0;
		double over = 1.01;
		if(t < 0.1){
			siki = 0.99;
		}
		else{
			if(etest >= 0.01){
				if(t < 0.8) siki = -0.2 * t + 1;
			}
			else if(avrutil > 0.5){
				double gap = maxutil - minutil;
				siki = avrutil + gap;
				over = siki + 0.1;
				if(siki >1.01) siki = 0.99;
			}
			else{
				double gap = maxutil - minutil;
				siki = 0.9 - gap * t; 
				over = siki + 0.1;
			}
		}
			while(scr < siki || scr > over) {
				bid = getRandomBid();
				scr = getUtility(bid)/Math.pow(discountFactor, timeline.getTime());
			}
			bst = bid; 
		return bst;
	}
		else
	{
		return(maxbid);
	}
	}
	

	
	
	/**
	 * This function determines the accept probability for an offer.
	 * At t=0 it will prefer high-utility offers.
	 * As t gets closer to 1, it will accept lower utility offers with increasing probability.
	 * it will never accept offers with utility 0.
	 * @param u is the utility 
	 * @param t is the time as fraction of the total available time 
	 * (t=0 at start, and t=1 at end time)
	 * @return the probability of an accept at time t
	 * @throws Exception if you use wrong values for u or t.
	 * 
	 */
	double Paccept(double u, double t1) throws Exception
	{
		double t=t1*t1*t1; // steeper increase when deadline approaches.
		if (u<0 || u>1.05) throw new Exception("utility "+u+" outside [0,1]");
		// normalization may be slightly off, therefore we have a broad boundary up to 1.05
		if (t<0 || t>1) throw new Exception("time "+t+" outside [0,1]");
		if (u>1.) u=1;
		if (t==0.5) return u;
		return (u - 2.*u*t + 2.*(-1. + t + Math.sqrt(sq(-1. + t) + u*(-1. + 2*t))))/(-1. + 2*t);
	}

	double sq(double x) { return x*x; }
}
