//package agents;
import java.math.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.util.Iterator;
import negotiator.Agent;
import negotiator.Bid;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.EndNegotiation;
import negotiator.actions.Offer;
import negotiator.issue.Issue;
import negotiator.issue.ISSUETYPE;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.IssueInteger;
import negotiator.issue.IssueReal;
import negotiator.issue.Value;
import negotiator.issue.ValueDiscrete;
import negotiator.issue.ValueInteger;
import negotiator.issue.ValueReal;
import negotiator.utility.Evaluator;
import negotiator.utility.EVALUATORTYPE;
import negotiator.utility.EvaluatorDiscrete;
/**
 * @author W.Pasman
 * Some improvements over the standard SimpleAgent.
 */
public class AgentLinear extends Agent
{
	private Action actionOfPartner=null;
	private static double MINIMUM_BID_UTILITY = 0.5;
	private ArrayList<Bid> bids=new ArrayList<Bid>();
	private ArrayList<Bid> sortedBids=new ArrayList<Bid>();
	private ArrayList<Bid> oponentBids=new ArrayList<Bid>();
	Bid maxBid = null;
	double hardHeadTime=0.8;
	double acceptAnythingTime=0.95;
	double minimumAcceptedUtility=0.7;
	double panicTimeUtility=minimumAcceptedUtility;
	double m=(double)(1.0-minimumAcceptedUtility)/(double)(hardHeadTime-acceptAnythingTime);
    double b=(double)((hardHeadTime*minimumAcceptedUtility)-acceptAnythingTime)/(double)(hardHeadTime - acceptAnythingTime);	
	int noOfRounds=0;
	double reservationValue=0.0;
	double discountFactor=1.0;
	boolean discountFactorAle=false;
	boolean panicTimeUtilityAdjusted=false;
	double minimumAcceptedPanicTimeUtility=0.6;
	/**
	 * init is called when a next session starts with the same opponent.
	 */
	public void init()
	{
		
		/*
		 * Concerns:
		 * I have not used cost when calculating utility , but sometimes cost is null
		 *  evaluation of an issue value can not be 0.
		 *  new Bid () throws an exception
		 * cost can be null
		 * I can use Value.getValue(issueNumber) to get the value
		 * becareful of Value being changed to com.sun.jdi.Value
		 * reservation value is null for domains with no reservation value
		 *make ArrayList<issues> issues global
		 *since the last step could take time assume negotiation ends at 0.9t
		 *look for a method to estimate the middle utility
		 */
		
		//System.out.println("My Utility Space");
		try {
			
			this.discountFactor=utilitySpace.getDiscountFactor();
			if (discountFactor==0.0){
				discountFactor=1;
			}
			//System.out.println("Discount Factor"+utilitySpace.getDiscountFactor());
			hardHeadTime=hardHeadTime*discountFactor;
			acceptAnythingTime=acceptAnythingTime*discountFactor;
		
		}
		catch(Exception ex) {
			//System.out.println("error when seeting DF");
		}
		
    try {
			
			//System.out.println("Reservation Value"+utilitySpace.getReservationValue());
			this.reservationValue=utilitySpace.getReservationValue();
			
			//this.discountFactor=utilitySpace.getDiscountFactor();
			//System.out.println("Discount Factor"+utilitySpace.getDiscountFactor());
		}
		catch(Exception ex) {
			//System.out.println("error when seeting RV");
		}
		
		ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
		
		//for (Bid bid:bids){
			/*
			for (Issue issue:issues){
				try
				{
					Value value=bid.getValue(issue.getNumber());
					System.out.println(value);
			    }
				catch (Exception ex){
					System.out.println("error when listing bids");
				}
	      }
			*/
			
			//System.out.println(getUtility(bid));
		//}
		

	
		//ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();	
		//System.out.println("no of issues:"+ issues.size());
		//HashMap<Integer,Value> values = new HashMap<Integer, Value>();
		HashMap<Integer,ArrayList<negotiator.issue.Value>> theValues = new HashMap<Integer,ArrayList<negotiator.issue.Value>>();
		
		try {
	    	 // System.out.println("discount factor"+utilitySpace.getDiscountFactor());
	    	 // System.out.println("reservation value "+ utilitySpace.getReservationValue());
		for (Issue issue:issues){
			Evaluator eval= utilitySpace.getEvaluator(issue.getNumber());
			Integer issueNo=issue.getNumber();
			ArrayList<negotiator.issue.Value> values =new ArrayList<negotiator.issue.Value>();
			switch(issue.getType()){
			case DISCRETE:
			      EvaluatorDiscrete evalDiscrete=(EvaluatorDiscrete)eval;
			      IssueDiscrete issueDiscrete=(IssueDiscrete)issue;
			      //System.out.println("issue "+ issue.getNumber() + "weight" + evalDiscrete.getWeight());
			      //System.out.println("Max Value " + evalDiscrete.getMaxValue() +"evaluation" + evalDiscrete.getEvaluation((ValueDiscrete)evalDiscrete.getMaxValue()) );
			      int noOfValues=issueDiscrete.getNumberOfValues();
                  //values.put(issue.getNumber(),(Value)evalDiscrete.getMaxValue());
			      for (int i=0;i<noOfValues;i++){
                  //System.out.print(issueDiscrete.getValue(i));
                  try
                  {
                  //System.out.print(evalDiscrete.getEvaluation(issueDiscrete.getValue(i))); 
                  //System.out.print("Cost: ");
                  //System.out.println(evalDiscrete.getCost(issueDiscrete.getValue(i)));
                  values.add((Value)issueDiscrete.getValue(i));
  
                  }
                  catch (Exception e){
                	  //System.out.println("error occured during evaluation");
                  }
            }
			      break;
			case REAL: 
				try {
					IssueReal issueReal = (IssueReal)issue;
					//System.out.println("Real Issue");
					//System.out.println("Discr."+ issueReal.getNumberOfDiscretizationSteps());
					//System.out.println("LB"+issueReal.getLowerBound());
					//System.out.println("UB"+issueReal.getUpperBound());
					double oneStep=(double)(issueReal.getUpperBound()-issueReal.getLowerBound())/(double)issueReal.getNumberOfDiscretizationSteps();
					for (int j=0;j<=issueReal.getNumberOfDiscretizationSteps();j++){
                            //System.out.println(issueReal.getLowerBound() + (double)j*oneStep);
                            values.add((Value)(new ValueReal(issueReal.getLowerBound() + (double)j*oneStep)));
                     }
				}
				catch(Exception ex){
					//System.out.println("error in iterating real issues");
				}
				
				//int optionInd = randomnr.nextInt(lIssueReal.getNumberOfDiscretizationSteps()-1);
				//values.put(lIssueReal.getNumber(), new ValueReal(lIssueReal.getLowerBound() + (lIssueReal.getUpperBound()-lIssueReal.getLowerBound())*(double)(optionInd)/(double)(lIssueReal.getNumberOfDiscretizationSteps())));
			
			}
			
			
			
			
			//bidByConceding ();
			theValues.put(issueNo,values);		
		}
		//maxBid=new Bid(utilitySpace.getDomain(),values);
		int maxNoOfValues=0;
		int maxNoOfBids=1;
		Iterator it = theValues.keySet().iterator();
		while (it.hasNext()){
			Integer issueNumber =(Integer) it.next();
			ArrayList<negotiator.issue.Value> values =theValues.get(issueNumber);
			maxNoOfBids=maxNoOfBids*values.size();
			if (values.size() > maxNoOfValues){
				  maxNoOfValues=values.size();
			}
			
		}
		if (maxNoOfBids > 10000){
			generateNonExhaustiveBids (theValues,maxNoOfValues);	
		}
		else {
			generateBids();
		}
		//for (Bid bid:sortedBids){
        //System.out.println(getUtility(bid));
	    //}
	    //sleep(1);
		//generateBids ();
		//System.out.println("total no of bids "+ sortedBids.size());
		}
		catch(Exception ex) 
	    {
	     //System.out.println("error in iterating issues");	
	    }
		
		
		//*/
		if(utilitySpace.getReservationValue() != null)
			MINIMUM_BID_UTILITY = utilitySpace.getReservationValue();
		   
	}
	
	
	

	public static String getVersion() { 
		
		return "1"; 
		
	}
	
	@Override
	public String getName()
	{
		
		return "AgentLinear";
	}

	
	private void generateNonExhaustiveBids (HashMap<Integer,ArrayList<negotiator.issue.Value>> theValues, int maxNoOfValues){
		
		
		ArrayList<Bid> nonExhaustiveBids=new ArrayList<Bid>();
		for (int j=0;j<maxNoOfValues;j++){
			Iterator it = theValues.keySet().iterator();
			HashMap<Integer,negotiator.issue.Value> valuesForBid=new HashMap<Integer,negotiator.issue.Value>();
			while (it.hasNext()){
				Integer issueNumber =(Integer) it.next();
				ArrayList<negotiator.issue.Value> values =theValues.get(issueNumber);
				if (j < values.size()){
					valuesForBid.put(issueNumber,values.get(j));
				}
				else {
					valuesForBid.put(issueNumber,values.get(0));
				}
				
			}
			try{
			nonExhaustiveBids.add(new Bid(utilitySpace.getDomain(),valuesForBid));
			}
			catch (Exception ex){
				//continue
			}
			
		}
		sortedBids=sort(nonExhaustiveBids);
	}
	
	private boolean generateBids (){
		
		HashMap<Integer,negotiator.issue.Value> values = new HashMap<Integer, negotiator.issue.Value>();
		ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
		int noOfIssues=issues.size();
		int curI=0;
		boolean success=recursiveGenerateBids(issues,curI,noOfIssues,values);
		//System.out.println("in generate bids");
		//for (Bid bid:bids){
          //     System.out.println(getUtility(bid));
		//}
		
		//System.out.println("sortedBids");
		sortedBids=sort(bids);
		//for (Bid bid:sortedBids){
            //System.out.println(getUtility(bid));
		//}
		//sleep(0.5);
		//System.out.println("after sleep");
		//ArrayList<Bid> sortedBids=sort(bids);
		//for (Bid bid:sortedBids){
          //  System.out.println(getUtility(bid));
		//}
		
		return true;
	}
	
	private boolean recursiveGenerateBids(ArrayList<Issue> issues, int curI,int noOfIssues,HashMap<Integer,negotiator.issue.Value> values){

		if (curI == noOfIssues){
		    try {
		    
		    HashMap<Integer,negotiator.issue.Value> valuesForBid=new HashMap<Integer,negotiator.issue.Value>();
		    valuesForBid=(HashMap<Integer,negotiator.issue.Value>)values.clone();
		    Iterator it = valuesForBid.keySet().iterator();
		   // while(it.hasNext()){
		    	Integer issueNumber = (Integer)it.next();
		    	//System.out.println(issueNumber+ "issue when making bids" + valuesForBid.get(issueNumber));
		   // }
		   
		    Bid aBid=new Bid(utilitySpace.getDomain(),valuesForBid);
		   // System.out.println(getUtility(aBid));
		    bids.add(aBid);
		    //for (Bid bid:bids){
		    	//System.out.println(getUtility(bid));
		    //}

		    }
		    catch(Exception ex) 
		    {
		    	//System.out.println("error in when making bids");
		    }
			return true;              	
		}
		
		if (curI<noOfIssues){
		    Issue curIssue=issues.get(curI);
		    ISSUETYPE type=curIssue.getType();
		   if (type==ISSUETYPE.DISCRETE){
			    IssueDiscrete issueDiscrete=(IssueDiscrete)curIssue;
			    int noOfValues=issueDiscrete.getNumberOfValues();
			    for (int i=0;i<noOfValues;i++){
			    //System.out.println(curIssue.getNumber()+"issue"+ (Value)issueDiscrete.getValue(i) );
			    values.put(curIssue.getNumber(),(Value)issueDiscrete.getValue(i));
			    recursiveGenerateBids(issues, (curI+1),noOfIssues,values);	
			    }
		   }
		  
           if (type==ISSUETYPE.REAL){
        	    IssueReal issueReal=(IssueReal)curIssue;
			    int noOfValues=issueReal.getNumberOfDiscretizationSteps();
			    //System.out.println("Real Issue");
				//System.out.println("Discr."+ issueReal.getNumberOfDiscretizationSteps());
				//System.out.println("LB"+issueReal.getLowerBound());
				//System.out.println("UB"+issueReal.getUpperBound());
				double oneStep=(double)(issueReal.getUpperBound()-issueReal.getLowerBound())/(double)issueReal.getNumberOfDiscretizationSteps();
				for (int j=0;j<=issueReal.getNumberOfDiscretizationSteps();j++){

					 //System.out.println(issueReal.getLowerBound() + (double)j*oneStep);
					 Value realValue=(Value)(new ValueReal(issueReal.getLowerBound() + (double)j*oneStep));
					 values.put(curIssue.getNumber(),realValue);
					 recursiveGenerateBids(issues, (curI+1),noOfIssues,values);
				}
			  
		   }
		   
           
		    
		 	 
		}
		
		return true;
		
	}
	
	private Bid bidByConceding (double targetUtility){
    //double time=timeline.getTime();
	
    	//System.out.println("time"+ time +"target utility "+ targetUtility);
    	
    	//double targetUtility=m*time+b;
    	//Bid aBid;
    	int lastIndex=-1;
        for (Bid bid:sortedBids){    
        	if (getUtility(bid) >= targetUtility){
        		
        		lastIndex=lastIndex+1;
            }
        	else {
        		break;
        	}
        
        //System.out.print("time"+ time +"target utility ");
    	//System.out.println (targetUtility);
    	//System.out.println("bid util."+ getUtility(sortedBids.get(lastIndex)));
    	//time=timeline.getTime();
    }
        
        Random randomnr= new Random();
    	
	 // return  sortedBids.get(randomnr.nextInt(lastIndex+1));	
		if (lastIndex <0){
			lastIndex=0;
		}
        
		
		return sortedBids.get(randomnr.nextInt(lastIndex+1));
	}
	
	private Bid copyBid (Bid aBid){
		   
		    Bid copyBid=null;
		    try {
		        ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
				HashMap<Integer,negotiator.issue.Value> values = new HashMap<Integer, negotiator.issue.Value>();
		        for (Issue issue:issues){
		        	values.put(issue.getNumber(), aBid.getValue(issue.getNumber()));
		        }
		        copyBid=new Bid (utilitySpace.getDomain(),values);
		
			}
		    catch(Exception ex) {
		    	
		    	
		    	//throw ex;
		
		    }
		    return copyBid;
			
	}
	public void ReceiveMessage(Action opponentAction) 
	{
		//System.out.println("RaiyeAgent");
		//System.out.println(utilitySpace.discountFactor);
		/*
		ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
		System.out.println("no of issues:"+ issues.size());
		for (Issue issue:issues){
			
			System.out.println("issue type"+ issue.getType());
	 		switch (issue.getType()){
			case DISCRETE:
				IssueDiscrete issueDiscrete=(IssueDiscrete)issue;
				System.out.println("issue " + issue.getNumber());
                for (int i=0;i<issueDiscrete.getNumberOfValues();i++){
                       System.out.println(issueDiscrete.getValue(i));                	
                }

			}
			
		}
		*/
		actionOfPartner = opponentAction;
	}


	public Action chooseAction()
	{
		Action action = null;
      
		 
		
		try 
		{ 
			
			
			 if(actionOfPartner==null)  {
					Bid maxBid=sortedBids.get(0);
					//System.out.println("My Max Bid"+ getUtility(maxBid));
				   
	                Bid maxBidForOffer=copyBid(maxBid);
					action=new Offer(getAgentID(), maxBidForOffer);
	                
	    			//Bid maxBidClone
					//action=new Offer(getAgentID(), maxBid);
				}
			 
			
               
				
				
				
			  
				if (actionOfPartner instanceof Offer)
				{ 
					boolean acceptedBeforeSanityCheck=false;
					boolean accepted=false;
					double time = timeline.getTime();
		           			 
					Bid oponentBid=((Offer)(actionOfPartner)).getBid();
					Bid copyOfOponentBid=copyBid(oponentBid);
					ArrayList<Bid> temporaryOponentBids=sortedAdd(copyOfOponentBid,oponentBids);
					oponentBids=temporaryOponentBids;
					double oppnentBidUtility=getUtility(copyOfOponentBid);
					double discountedPanicTimeUtility=panicTimeUtility*Math.pow(discountFactor,time);
					double maxUtility= 1.0*Math.pow(discountFactor,time);
					/*
					if ( time > acceptAnythingTime){
						
						if ((oppnentBidUtility-discountedPanicTimeUtility) >=0){
							 action=new Accept(getAgentID());
						     return action;
						 }
						 
					 }
					 else 
					{
						 noOfRounds=noOfRounds+1;
						 System.out.println("Round"+noOfRounds);
					}
					*/
					//for (Bid bid:oponentBids){
						//System.out.println(getUtility(bid));
					//}
	                
					
					if ( Math.abs((maxUtility - oppnentBidUtility))<=.01){
	                	action=new Accept(getAgentID()); 
	                	
	                	acceptedBeforeSanityCheck=true;
	                	//System.out.println("he gave me 1.0");
	                }
	                
					if (acceptedBeforeSanityCheck==false){
						Bid maxBid=sortedBids.get(0);
						Bid maxBidForOffer=copyBid(maxBid);
						//double time=timeline.getTime();
						Bid previousOponentMaxBid=null;
		                if (oponentBids.size()>0){
		                	previousOponentMaxBid=oponentBids.get(0);
		                }
						double targetUtility;
						//no discount
						if (time < hardHeadTime){
						  
                        	targetUtility=1.0;	
						
                        }
                        else
                        {
                        
                             targetUtility=m*(time)+b;
						
					     }
						
						
                        
                        
                        if (time > acceptAnythingTime){
			                 
                        	
                        	 if (panicTimeUtilityAdjusted==false){
                        		 
                        		 if (previousOponentMaxBid!=null){
                                      
                        			 double undiscountedPreviousOponentBidUtility=getUtility(previousOponentMaxBid)/(Math.pow(discountFactor, time));
                        			 double difference=panicTimeUtility -undiscountedPreviousOponentBidUtility;
                        			 if (difference > 0)
                        			 {
                        			 panicTimeUtility= panicTimeUtility - difference/2.0;
                        			 }
                        			 if(panicTimeUtility < minimumAcceptedPanicTimeUtility){
                        				 panicTimeUtility= minimumAcceptedPanicTimeUtility;           				 
                        			 }
                        			 
                        			 
                        		 }		 
                        	 
                        		 panicTimeUtilityAdjusted=true;
                        	 }
							 targetUtility =panicTimeUtility;
						}
						//double targetUtility=m*(time+Math.pow(time, discountFactor))+b;
						
                        
                        targetUtility=targetUtility*Math.pow(discountFactor,time);
                        
                        //System.out.println("df"+discountFactor);
                        //System.out.println("maxUtil"+maxUtility);
                        //System.out.println("panicUtil"+discountedPanicTimeUtility);
    				    //System.out.println("targetUtil"+targetUtility);
    				    //System.out.println("getDF"+ utilitySpace.getDiscountFactor());
						

    				    Bid concendedBid=bidByConceding (targetUtility);
		                Bid concendedBidForOffer=copyBid(concendedBid);
		               
						Bid offerBid;
		                Random randomnr= new Random();
		                if (randomnr.nextInt(2)>-1){
		                	offerBid=concendedBidForOffer;
		                }
		                else
		                {
		                	offerBid=maxBidForOffer;
		                }
		               
		                
		                
		                if (previousOponentMaxBid!=null){
		                	
		                	if(getUtility(previousOponentMaxBid) > getUtility(concendedBidForOffer)){
		                             offerBid=copyBid(previousOponentMaxBid);            		
		                	}
		                }
		                
		                if (Math.abs (getUtility(copyOfOponentBid) - getUtility(offerBid))<0.01){
		                	
		                	accepted=true;
		                }
		                
		                
		                double discountedReservationValue=reservationValue*Math.pow(discountFactor, time);
		                if (accepted==true){
		                	
		                	if (getUtility(copyOfOponentBid) < discountedReservationValue){
		                		action=new EndNegotiation(getAgentID());                		
		                	}
		                	else
		                	{
		                	action=new Accept(getAgentID()); 
		                	}
		                }
		                else{
		                
		                	if (getUtility(offerBid) < discountedReservationValue){
		                		action=new EndNegotiation(getAgentID());                		
		                    }
		                	else
		                	{
		                		action=new Offer(getAgentID(), offerBid);
		                	
		                	}
		                	
		                	
		                	
		                }
		                
		                
		                //action=new Offer(getAgentID(), offerBid);
					
					
					}
			}
			//sleep(0.005); // just for fun
		} catch (Exception e) { 
			//System.out.println("Exception in ChooseAction:"+e.getMessage());
			action=new Accept(getAgentID()); // best guess if things go wrong. 
		}
		return action;
	}

	private double calculateUtility(HashMap<Integer,negotiator.issue.Value> values){
		Double utility=0.0;
		Double totalCost=0.0;
		//System.out.println("in calaculate utility");
		Iterator it = values.keySet().iterator();
		while (it.hasNext()) {
			Integer issueNumber =(Integer) it.next();
		    double evaluation=0.0;
			double weight=0.0;
			double cost=0.0;
			Evaluator evaluator = utilitySpace.getEvaluator((int)issueNumber);
		    weight=evaluator.getWeight();
		    switch (evaluator.getType()){
		    case DISCRETE:
		        EvaluatorDiscrete evaluateDiscrete=(EvaluatorDiscrete) evaluator;
		        ValueDiscrete valueDiscrete=(ValueDiscrete)values.get(issueNumber);
		        try{
		        evaluation=evaluateDiscrete.getEvaluation(valueDiscrete);
		        cost=evaluateDiscrete.getCost(valueDiscrete);
		        }
		        catch (Exception ex)
		        {
		        	//System.out.println("Exception in calculate utility "+ex.toString());
		        }
		       }
		    
		    utility=utility+(evaluation*weight);
		    totalCost=totalCost+cost;
		    }  
     	    
		
		    return utility;
	}
	private double calculateDiscountedUtility(HashMap<Integer,negotiator.issue.Value> values)
	{
	 double utility=0.0;
	 double discountFactor=utilitySpace.getDiscountFactor();
     double utilityWithoutDiscount=calculateUtility(values);
     double time = timeline.getTime();
     if (discountFactor > 0){
    	 
    	 utility=utilityWithoutDiscount*(Math.pow(discountFactor,time));
     }
     utility=utilityWithoutDiscount;
     return utility;
	}
	private boolean isAcceptable(double offeredUtilFromOpponent, double myOfferedUtil, double time) throws Exception
	{
		double P = Paccept(offeredUtilFromOpponent,time);
		if (P > Math.random())
			return true;		
		return false;
	}

	/**
	 * Wrapper for getRandomBid, for convenience.
	 * @return new Action(Bid(..)), with bid utility > MINIMUM_BID_UTIL.
	 * If a problem occurs, it returns an Accept() action.
	 */
	private Action chooseRandomBidAction() 
	{
		Bid nextBid=null ;
		try { nextBid = getRandomBid(); }
		catch (Exception e) { 
	//		System.out.println("Problem with received bid:"+e.getMessage()+". cancelling bidding"); 
			
		}
		if (nextBid == null) return (new Accept(getAgentID()));                
		return (new Offer(getAgentID(), nextBid));
	}

	/**
	 * @return a random bid with high enough utility value.
	 * @throws Exception if we can't compute the utility (eg no evaluators have been set)
	 * or when other evaluators than a DiscreteEvaluator are present in the util space.
	 */
	private Bid getRandomBid() throws Exception
	{
		HashMap<Integer, Value> values = new HashMap<Integer, Value>(); // pairs <issuenumber,chosen value string>
		ArrayList<Issue> issues=utilitySpace.getDomain().getIssues();
		Random randomnr= new Random();

		// create a random bid with utility>MINIMUM_BID_UTIL.
		// note that this may never succeed if you set MINIMUM too high!!!
		// in that case we will search for a bid till the time is up (2 minutes)
		// but this is just a simple agent.
		Bid bid=null;
		do 
		{
			for(Issue lIssue:issues) 
			{
				switch(lIssue.getType()) {
				case DISCRETE:
					IssueDiscrete lIssueDiscrete = (IssueDiscrete)lIssue;
					int optionIndex=randomnr.nextInt(lIssueDiscrete.getNumberOfValues());
					values.put(lIssue.getNumber(), lIssueDiscrete.getValue(optionIndex));
					break;
				case REAL:
					IssueReal lIssueReal = (IssueReal)lIssue;
					int optionInd = randomnr.nextInt(lIssueReal.getNumberOfDiscretizationSteps()-1);
					values.put(lIssueReal.getNumber(), new ValueReal(lIssueReal.getLowerBound() + (lIssueReal.getUpperBound()-lIssueReal.getLowerBound())*(double)(optionInd)/(double)(lIssueReal.getNumberOfDiscretizationSteps())));
					break;
				case INTEGER:
					IssueInteger lIssueInteger = (IssueInteger)lIssue;
					int optionIndex2 = lIssueInteger.getLowerBound() + randomnr.nextInt(lIssueInteger.getUpperBound()-lIssueInteger.getLowerBound());
					values.put(lIssueInteger.getNumber(), new ValueInteger(optionIndex2));
					break;
				default: throw new Exception("issue type "+lIssue.getType()+" not supported by SimpleAgent2");
				}
			}
			bid=new Bid(utilitySpace.getDomain(),values);
		} while (getUtility(bid) < MINIMUM_BID_UTILITY);

		return bid;
	}

	/**
	 * This function determines the accept probability for an offer.
	 * At t=0 it will prefer high-utility offers.
	 * As t gets closer to 1, it will accept lower utility offers with increasing probability.
	 * it will never accept offers with utility 0.
	 * @param u is the utility 
	 * @param t is the time as fraction of the total available time 
	 * (t=0 at start, and t=1 at end time)
	 * @return the probability of an accept at time t
	 * @throws Exception if you use wrong values for u or t.
	 * 
	 */
	
	
  /*
	ArrayList<negotiator.issue.Value> sortDiscreteValues (EvaluatorDiscrete evalDiscrete,ArrayList<negotiator.issue.negotiator.issue.ValueDiscrete> discreteValues){
	   ArrayList<negotiator.issue.negotiator.issue.ValueDiscrete> sortedDValues=new ArrayList<negotiator.issue.negotiator.issue.ValueDiscrete>();
	   for (negotiator.issue.negotiator.issue.ValueDiscrete dValue:discreteValues){
		    double dValueEvaluation=evalDiscrete.getEvaluation(dValue);
			//System.out.println(bidUtility);
		    boolean inserted =false;
		    for (int i=0;i<sortedDValues.size();i++){
		    	//System.out.print("in for");
		    	if (dValue > getUtility(sortedBidsList.get(i))){
		                  sortedBidsList.add(i,bid);	
		                 // System.out.println("inserted");
		                  inserted=true;
		                  break;
				}
				
			}
		    if (inserted!=true){
				//System.out.println("first inserted");
				sortedBidsList.add(bid);
			}
			
			
		}
		
	   
	   
	   
   }
   */
   ArrayList<Bid> sort (ArrayList<Bid> bids){
		ArrayList<Bid> sortedBidsList =new ArrayList<Bid>();
		//System.out.println("in sort bids");
		
		for (Bid bid:bids){
		    double bidUtility=getUtility(bid);
			//System.out.println(bidUtility);
		    boolean inserted =false;
		    for (int i=0;i<sortedBidsList.size();i++){
		    	//System.out.print("in for");
		    	if (bidUtility > getUtility(sortedBidsList.get(i))){
		                  sortedBidsList.add(i,bid);	
		                 // System.out.println("inserted");
		                  inserted=true;
		                  break;
				}
				
			}
		    if (inserted!=true){
				//System.out.println("first inserted");
				sortedBidsList.add(bid);
			}
			
			
		}
		
		return sortedBidsList;
	}
	ArrayList<Bid> sortedAdd (Bid bid,ArrayList<Bid> bids){

		
		double newBidUtility=getUtility(bid);
		ArrayList<Bid> sortedBidsList=new ArrayList<Bid>();
		if (bids.size()==0){
			sortedBidsList.add(bid);
		}
		int curIndex=0;
		for (Bid aBid:bids){
            
			if ((Math.abs(newBidUtility - getUtility(aBid))) < 0.01){
				 sortedBidsList.add(aBid);
                 curIndex=curIndex+1;
				 //curIndex=curIndex+1;
				break;
			}
			
			if (newBidUtility < getUtility(aBid) ){
                  //sortedBids.add(bid);
                  sortedBidsList.add(aBid);
                  curIndex=curIndex+1;
  			}
			else 
			{
				sortedBidsList.add(bid);
				break;
				
			}
	     		
		}
		for (int i=curIndex;i<bids.size();i++){
			sortedBidsList.add(bids.get(i));
		}
		
		
		
		return sortedBidsList;
	}
    double Paccept(double u, double t1) throws Exception
	{
		double t=t1*t1*t1; // steeper increase when deadline approaches.
		if (u<0 || u>1.05) throw new Exception("utility "+u+" outside [0,1]");
		// normalization may be slightly off, therefore we have a broad boundary up to 1.05
		if (t<0 || t>1) throw new Exception("time "+t+" outside [0,1]");
		if (u>1.) u=1;
		if (t==0.5) return u;
		return (u - 2.*u*t + 2.*(-1. + t + Math.sqrt(sq(-1. + t) + u*(-1. + 2*t))))/(-1. + 2*t);
	}

	double sq(double x) { return x*x; }
}
