package Rumba;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.SortedMap;
import java.util.TreeMap;

import negotiator.Bid;
import negotiator.BidIterator;
import negotiator.utility.UtilitySpace;

public class BidSpace {
	
	private List<EvaluatedBid> evaluatedBids;
	private TreeMap<Double, EvaluatedBid> evaluatedBidsMap;
	private Random random;
		
	public BidSpace(UtilitySpace utilitySpace)
	{
		this.random = new Random();
				
	    initEvaluatedBids(utilitySpace);
	}
	
	public SortedMap<Double, EvaluatedBid> getEvaluatedBids() {
		return evaluatedBidsMap;
	}
	
	public EvaluatedBid randomBid() {
		int index = this.random.nextInt(this.evaluatedBids.size());
		return this.evaluatedBids.get(index);
	}
		
	public EvaluatedBid maxBid() {
		return extractBid(this.evaluatedBidsMap.lastEntry());
	}
	
	public EvaluatedBid lowerBid(double utility) {
		return extractBid(this.evaluatedBidsMap.lowerEntry(utility));
	}
	
    public EvaluatedBid lowerBid(EvaluatedBid bid) {
    	return lowerBid(bid.getUtility());
    }
	
	public EvaluatedBid floorBid(double utility) {
		return extractBid(this.evaluatedBidsMap.floorEntry(utility));
	}
	
    public EvaluatedBid floorBid(EvaluatedBid bid) {
    	return floorBid(bid.getUtility());
    }
		
	private EvaluatedBid extractBid(Map.Entry<Double, EvaluatedBid> entry) {
		if (entry != null) {
			return entry.getValue();
		}
		else {
			return null;
		}
	}
	
	private void initEvaluatedBids(UtilitySpace utilitySpace) {
			
		// Generate all bids and evaluated them
		this.evaluatedBids = new ArrayList<EvaluatedBid>();
		
		BidIterator bidIterator = new BidIterator(utilitySpace.getDomain());
		while (bidIterator.hasNext()) {
	    	try {
	    		Bid bid = bidIterator.next();
	    		double utility = utilitySpace.getUtility(bid);
	    		EvaluatedBid evaluatedBid = new EvaluatedBid(bid, utility);
	    		this.evaluatedBids.add(evaluatedBid);
	    	} catch (Exception e) {
	    		//System.out.println(e.getStackTrace());
	    	}
		}
	    
	    this.evaluatedBidsMap = new TreeMap<Double, EvaluatedBid>();
	    for (EvaluatedBid evaluatedBid : this.evaluatedBids) {
	    	this.evaluatedBidsMap.put(evaluatedBid.getUtility(), evaluatedBid);
	    }
	}
}
