package Rumba;

import negotiator.Agent;
import negotiator.actions.Action;
import negotiator.actions.Offer;

/** Main agent class for Rumba
 * 
 * @author Shahar Kosti
 * @author Mor Vered
 */
public class Rumba extends Agent {
	
	private BidSpace bidSpace;
	private Logic logic;
	private ThresholdCalculator thresholdCalculator;
		
	public void init()
	{
		this.bidSpace = new BidSpace(utilitySpace);
		initThresholdCalculator();
		this.logic = new Logic(this, this.utilitySpace, 
				this.bidSpace, this.thresholdCalculator);
	}

	public static String getVersion() { 
		return "1.2"; 
	}
	
	public String getName() {
		return "Rumba";
	}

	public void ReceiveMessage(Action action) {
		if (!(action instanceof Offer)) {
			return;
		}
				
		Offer offer = (Offer)action;
		this.logic.setOpponentOffer(offer);
	}	
	
	public Action chooseAction() {
		return this.logic.chooseAction();
	}
	
	private void initThresholdCalculator() {
		double maxUtil = 1;
		double minUtil = 0.585;
		
		try
		{
			EvaluatedBid maxBid = this.bidSpace.maxBid();
			maxUtil = maxBid.getUtility();
			
			if(utilitySpace.getReservationValue() != null) {
				minUtil = utilitySpace.getReservationValue();
			}
		} catch (Exception e) {	
		}
		
		double discountFactor = this.utilitySpace.getDiscountFactor();
		
		this.thresholdCalculator = new ThresholdCalculator(
				this.timeline, discountFactor, minUtil, maxUtil);
	}
}
