package agents.OMAC_sp2012b;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.util.List;
import java.util.LinkedList;
import java.util.Collections;
import java.util.Date;
import java.util.Comparator;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Set;

import negotiator.Agent;
import negotiator.Bid;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.Offer;
import negotiator.issue.Issue;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.IssueInteger;
import negotiator.issue.IssueReal;
import negotiator.issue.Value;
import negotiator.issue.ValueDiscrete;
import negotiator.issue.ValueInteger;
import negotiator.issue.ValueReal;
import negotiator.Domain;
import negotiator.utility.Evaluator;
import negotiator.utility.UtilitySpace;
import negotiator.AgentID;

import java.io.BufferedWriter; 
import java.io.FileWriter; 
import java.io.File;
import java.io.Writer; 
import java.io.FileNotFoundException; 
import java.io.IOException; 
import java.io.*;

//ver. 1.06
class TimeBidHistory {	
	private List<Bid> fMyBids;
	private List<Bid> fOpponentBids;
	public List<Double> fTimes;  //new, record the time of receiving counter-offers
	
	public int curIndex = 0; //new, refer to the current position of opp bid history 
	public int curLength = 0; //new, the size of current bid window
	private double discount = 0;
	protected UtilitySpace fUtilitySpace;
	public double est_t =0;
	public double est_u =0;	
	public double maxU = 0.0;
	public int pMaxIndex = 0;
	public Bid bestOppBid=null;
	public double [] maxBlock; // record max
	public List<Integer> newMC;  // times where new maximum concession is given
	
	public TimeBidHistory(UtilitySpace pUtilitySpace, double dis) {
		discount = dis;
		fMyBids = new ArrayList<Bid>();	
		fTimes = new ArrayList<Double>();  //new	
		newMC = new ArrayList<Integer>();
		this.fUtilitySpace = pUtilitySpace;
				
		maxBlock = new double[100];
		for(int i=0; i < 100; i++){
			maxBlock[i] = 0.0;			
		}
						
	}
	
	/*
	 * Add our own bid to the list
	 */
	public void addMyBid(Bid pBid) {
		if (pBid == null)
			throw new IllegalArgumentException("pBid can't be null.");
		fMyBids.add(pBid);
				
	}
	
	/*
	 * returns the size (number) of offers already made
	 */
	public int getMyBidCount() {
		return fMyBids.size();
	}
	
	/*
	 * returns a bid from the list
	 */
	public Bid getMyBid(int pIndex) {
		return fMyBids.get(pIndex);
	}
	
	/*
	 * returns the last offer made
	 */
	public Bid getMyLastBid() {
		Bid result = null;
		if(getMyBidCount() > 0){
			result = fMyBids.get(getMyBidCount()-1);
		}
		return result;		
	}
	
	/*
	 * returns true if a bid has already been made before
	 */
	public boolean isInsideMyBids(Bid a){
		boolean result = false;
		for(int i = 0; i < getMyBidCount(); i++){
			if(a.equals(getMyBid(i))){
				result = true;
			}
		}
		return result;
	}
	
	/*
	 * add the bid the oppponent to his list
	 */
	public void addOpponentBidnTime(double oppU, Bid pBid, double time) {
		double undisOppU = oppU/Math.pow(discount,time);
		double nTime=time;
		ArrayList<Issue> lIssues = getIssues();	
		
		if (pBid == null)
			throw new IllegalArgumentException("vBid can't be null.");
		
		fTimes.add(time);
								
		if( undisOppU > maxU ){								
			maxU = undisOppU;  // prabably not useful in some sense
			//pMaxIndex = getOpponentBidCount()-1;
			pMaxIndex = fTimes.size()-1;
			bestOppBid = pBid;
			newMC.add(pMaxIndex);
		}
		
		if(nTime >=1.0)
			nTime = 0.99999;
			
		if ( maxBlock[ (int)Math.floor(nTime*100) ] < undisOppU  )
			maxBlock[ (int)Math.floor(nTime*100) ] = undisOppU;
			
	}
	
	public double[] getTimeBlockList(){
		return maxBlock;
	}
	
	/*
	 * returns the number of bids the opponent has made
	 */
	public int getOpponentBidCount() {
		return fOpponentBids.size();
	}
	
	/*
	 * returns the bid at a given index
	 */
	public Bid getOpponentBid(int pIndex) {
		return fOpponentBids.get(pIndex);
	}
	
	/*
	 * returns the opponents' last bid
	 */
	public Bid getOpponentLastBid() {
		Bid result = null;
		if(getOpponentBidCount() > 0){
			result = fOpponentBids.get(getOpponentBidCount()-1);
		}
		return result;
	}	
		
	public double getMyUtility(Bid b) {
		try {
			return this.fUtilitySpace.getUtility(b);
		} catch(Exception e) {			
			return 0;
		}		
	}
	
	public double getOpponentUtility(Bid b) {
		try {
			return this.fUtilitySpace.getUtility(b);
		} catch(Exception e) {		
			return 0;
		}		
	}
	
	/*
	 * returns the list of issues in the domain
	 */
	public ArrayList<Issue> getIssues() {
		return this.fUtilitySpace.getDomain().getIssues();
	}
	
	public double getFeaMC(double time) {
		int len = newMC.size();
		double dif = 1.0;
		
		if(len >= 3){
			dif = fTimes.get(newMC.get(len-1))- fTimes.get(newMC.get(len-3));
			dif = dif/2.0;			
		}else if(len >=2){
			dif = fTimes.get(newMC.get(len-1))- fTimes.get(newMC.get(len-2));						
		}else{
			dif = 0D;
		}
		
		//newmark("a-dif.txt",""+dif+","+len+","+time);
		return dif;	
	}
	
	public double getMCtime(){		
		if(newMC.size() == 0)
			return 0.0;
		
		return fTimes.get(newMC.get(newMC.size()-1));
	}
	
	
}
