package agents.gahboninho;



import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Random;
import java.util.TreeMap;


import negotiator.Agent;
import negotiator.AgentID;
import negotiator.Bid;
import negotiator.BidIterator;
import negotiator.Domain;
import negotiator.Timeline;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.Offer;
import negotiator.issue.*;
import negotiator.utility.UtilitySpace;

// The agent is similar to a bully.
// on the first few bids it steadily goes down to 0.9 of utility to make sure that if the opponent is trying to 
// profile me, he could do it more easily.
// after that, it goes  totally selfish and almost giving up no points.
// more over, the nicer the opponent is (Noise serves as a "niceness" estimation) the more selfish this agent gets.
// only at the last few seconds the agent panics and gives up his utility.
// I believe this is an excellent strategy, but sadly we did not have enough manpower and time
// to calibrate it and make it shine :(
// an opponent-model would have helped us improving the result on panic stages, and the noise(niceness) calculation is too rough.

public class Gahboninho extends Agent 
{
	
	final int PlayerCount = 10; // if player count is 10, then we
	// may give 10 points to opponent in order to give 1 point to ourselves
	
	class IssueManager
	{
		// world state :
		Timeline 	 		 T;
		UtilitySpace 		 US;
		int 				 TotalBiddingPossibilities; // amount of all possible bidsd
		// oponent properties :
		Bid					 OpponentBestBid; // the first bid opponent gave us
		double				 MyUtilityOnOpponentBestBid = 0;
		double 				 Noise = 0.4; // 0 to 1 - Tells us how varied the opponent's bids are, lately
		// My State:
		double 				 NextOfferUtility = 1;
		boolean 			 FirstOfferGiven = false;
		TreeMap<Double, Bid> Bids;
		HashMap<Issue, ArrayList<Value>> relevantValuesPerIssue = new HashMap<Issue, ArrayList<Value>>();
		Bid maxBid = null;
		
		
		public Bid GetMaxBidWithNoCost () throws Exception
		{
			Bid maxBid = this.US.getDomain().getRandomBid();
			Bid justBidding = this.US.getDomain().getRandomBid();
			
			
			for (Issue issue : this.US.getDomain().getIssues())
			{	
				double tmpUtil;
				double maxUtil = 0;
				int maxUtilValIndex = 0;
				
				switch(issue.getType()) 
				{
				
				case INTEGER:
					
					IssueInteger integerIssue =(IssueInteger)issue;
					
					justBidding.setValue(issue.getNumber(), new ValueInteger(integerIssue.getUpperBound()));
					maxUtil = US.getUtility(justBidding);
					
					justBidding.setValue(issue.getNumber(), new ValueInteger(integerIssue.getLowerBound()));
					tmpUtil = US.getUtility(justBidding);
					
					if (maxUtil > tmpUtil)
						maxBid.setValue(issue.getNumber(), new ValueInteger(integerIssue.getUpperBound()));
					else
						maxBid.setValue(issue.getNumber(), new ValueInteger(integerIssue.getLowerBound()));
					
					break;
					
				case REAL: 
					
					IssueReal realIssue =(IssueReal)issue;
					
					
					justBidding.setValue(issue.getNumber(), new ValueReal(realIssue.getUpperBound()));
					maxUtil = US.getUtility(justBidding);
					
					justBidding.setValue(issue.getNumber(), new ValueReal(realIssue.getLowerBound()));
					tmpUtil = US.getUtility(justBidding);
					
					if (maxUtil > tmpUtil)
						maxBid.setValue(issue.getNumber(), new ValueReal(realIssue.getUpperBound()));
					else
						maxBid.setValue(issue.getNumber(), new ValueReal(realIssue.getLowerBound()));

					break;
				case DISCRETE:
					IssueDiscrete discreteIssue = (IssueDiscrete)issue;
					int size = discreteIssue.getNumberOfValues();
					for (int i=0;i<size;i++)
					{
						justBidding.setValue(issue.getNumber(),discreteIssue.getValue(i) );
						tmpUtil = US.getUtility(justBidding);
						if (tmpUtil > maxUtil)
						{
							maxUtilValIndex = i;
							maxUtil = tmpUtil;
						}
					}
					
					maxBid.setValue(issue.getNumber(), discreteIssue.getValue(maxUtilValIndex));
					break;
				}
			}
			
			return maxBid;
		}
		// fill utility-to-bid map here:
		public IssueManager (UtilitySpace US, Timeline T)
		{
			this.T = T;
			this.US = US;
			try 
			{
				maxBid = GetMaxBidWithNoCost (); // try sparing the brute force
				double maxBidUtil = US.getUtility(maxBid);
				if ( maxBidUtil == 0) // in case cost comes into play
					this.maxBid = this.US.getMaxUtilityBid(); // use only if the simpler function won't work 
			
			} 
			catch (Exception e) 
			{
				try { this.maxBid = this.US.getMaxUtilityBid(); }
				catch (Exception e2){}
			}
			
			Bids  = new TreeMap<Double, Bid>();
			
			for (int i = 0; i < US.getDomain().getIssues().size(); ++i)
			{
				Issue I = (Issue) US.getDomain().getIssue(i);
				if (I.getType() == ISSUETYPE.DISCRETE)
				{
					IssueDiscrete ID = (IssueDiscrete)I;
					
					DifferentValuesCountPerIssueNum.put(ID.getNumber(), ID.getNumberOfValues() );
					OutgoingValueAppeareancesByIssueNum.put(ID.getNumber(), new TreeMap<String, Integer>());
				}
				else if (I.getType() == ISSUETYPE.REAL)
				{
					DifferentValuesCountPerIssueNum.put(I.getNumber(), (int) DiscretisationSteps);
					OutgoingValueAppeareancesByIssueNum.put(I.getNumber(), new TreeMap<String, Integer>());
				}
				else if (I.getType() == ISSUETYPE.INTEGER)
				{
					IssueInteger II = (IssueInteger)I;
					DifferentValuesCountPerIssueNum.put(I.getNumber(), 
							Math.min((int) DiscretisationSteps, 
									  II.getUpperBound() - II.getLowerBound() + 1 ) );
					
					OutgoingValueAppeareancesByIssueNum.put(I.getNumber(), new TreeMap<String, Integer>());
				}
			}
			ClearIncommingStatistics ();
		}
		void ClearIncommingStatistics ()
		{
			for (Issue I : US.getDomain().getIssues())
				IncomingValueAppeareancesByIssueNum.put(I.getNumber(), new TreeMap<String, Integer>());
			
		}
		public Bid getMaxBid()
		{
			return maxBid;
		}
		private double GetDiscountFactor ()
		{
			if (US.getDiscountFactor() <= 0.001 || US.getDiscountFactor() > 1)
				return 1;
			return US.getDiscountFactor();
		}	
		Random R = new Random();
		private void addPossibleValue(Issue issue, Value val)
		{
			if (!this.relevantValuesPerIssue.containsKey(issue))
			{
				this.relevantValuesPerIssue.put(issue, new ArrayList<Value>());
			}
			
			
			int randIndex = 0;
			if ( this.relevantValuesPerIssue.get(issue).size() > 0)
				randIndex = Math.abs(  R.nextInt() ) % this.relevantValuesPerIssue.get(issue).size();
			this.relevantValuesPerIssue.get(issue).add(randIndex,val);
		}

		final double DiscretisationSteps = 20; // minimum 2
		private void buildIssueValues(Bid firstOppBid) throws Exception
		{
			
			Bid justBidding = this.US.getDomain().getRandomBid();
			for (Issue issue : this.US.getDomain().getIssues())
			{
				justBidding.setValue(issue.getNumber(), firstOppBid.getValue(issue.getNumber()));
				double utilityWithOpp= this.US.getEvaluation(issue.getNumber(), justBidding);
				
				switch(issue.getType()) {
				case INTEGER:
					IssueInteger intIssue =(IssueInteger)issue;
					
					int iStep;
					int totalSteps= (int) Math.min(DiscretisationSteps - 1, intIssue.getUpperBound() - intIssue.getLowerBound() );
					iStep = Math.max(1, (int)((intIssue.getUpperBound() - intIssue.getLowerBound()) / totalSteps ));
					
					for (int i=intIssue.getLowerBound();i<=intIssue.getUpperBound();i+= iStep)
					{
						justBidding.setValue(issue.getNumber(), new ValueInteger(i));
						double utilityWithCurrent = this.US.getEvaluation(issue.getNumber(), justBidding);
						
						// Only see it as a possible value if it is better for us than the opponent offer
						if (utilityWithCurrent >= utilityWithOpp)
						{
							this.addPossibleValue(issue, new ValueInteger(i));
						}
					}
					break;
				case REAL: 
					
					IssueReal realIssue =(IssueReal)issue;
					double oneStep = (realIssue.getUpperBound()-realIssue.getLowerBound())/(DiscretisationSteps-1);
					for (double curr=realIssue.getLowerBound();curr<=realIssue.getUpperBound();curr+=oneStep)
					{
						justBidding.setValue(issue.getNumber(), new ValueReal(curr));
						double utilityWithCurrent = this.US.getEvaluation(issue.getNumber(), justBidding);
						// Only see it as a possible value if it is better for us than the opponent offer
						if (utilityWithCurrent >= utilityWithOpp)
						{
							this.addPossibleValue(issue, new ValueReal(curr));
						}
					}

					break;
				case DISCRETE:
					IssueDiscrete discreteIssue = (IssueDiscrete)issue;
					int size = discreteIssue.getNumberOfValues();
					for (int i=0;i<size;i++)
					{
						ValueDiscrete curr = discreteIssue.getValue(i);
						justBidding.setValue(issue.getNumber(), curr);
						double utilityWithCurrent = this.US.getEvaluation(issue.getNumber(), justBidding);
						// Only see it as a possible value if it is better for us than the opponent offer
						if (utilityWithCurrent >= utilityWithOpp)
						{
							this.addPossibleValue(issue, curr);
						}

					}
					break;
				}
			}
		}
		
		public void learnBids(Bid firstOppBid) throws Exception
		{
			this.buildIssueValues(firstOppBid);
			
			//CountdownToNoiseReestimation = Math.min(10, US.getDomain().getIssues().size());
			
			double startTime = T.getTime();
			
			Bid bid = this.US.getDomain().getRandomBid();
			
			// very hard to deal with hash map, so copy it to arraylist:
			Iterator<Entry<Issue,ArrayList<Value>>> MyIssueIterator = relevantValuesPerIssue.entrySet().iterator();
			while (MyIssueIterator.hasNext())
				IssueEntries.add(MyIssueIterator.next());
			
			// if there is a discount factor, don't take your time searching for bids
			BuildBid(new HashMap<Integer, Value>(),0, 0.1 * Math.pow( GetDiscountFactor(),0.5) + startTime);
		}
		private ArrayList<Entry<Issue,ArrayList<Value>>> IssueEntries = new ArrayList<Entry<Issue,ArrayList<Value>>>();
		private void BuildBid (HashMap<Integer,Value> B, int EntrySetIndex , double EndTime)
		{
			// TODO : build only bids with at least X util.
			// after some time, when we need lower utilities, re-build bid map, only this time consider
			// opponent's preferences: just build a method that tells us if one bid is preferrable (from opponent's point of view)
			// and if the less preferrable one gives us less utility than the other bid, simply remove the bid from map
			
			if ( T.getTime() < EndTime)
			{	
				if (EntrySetIndex < IssueEntries.size())
				{
					Entry<Issue, ArrayList<Value>> currentEntry = IssueEntries.get(EntrySetIndex);
					for (Value v : currentEntry.getValue())
					{
						B.put(currentEntry.getKey().getNumber(), v);
						BuildBid (B,EntrySetIndex + 1, EndTime);
					}
				}
				else
				{
					try 
					{
						
						Bid newBid = new Bid(US.getDomain(),(HashMap<Integer, Value>) B.clone() );
						
						Bids.put( US.getUtility(newBid), newBid );
						//System.out.print(US.getUtility(newBid));
						//System.out.println(" : " + B.toString());
					}
					catch (Exception e) 
					{
						//e.printStackTrace();
					}	
				}
				
			}
		}
		
		// returns the minimum utility we want in the next offer we send
		public double 	GetNextRecommendedOfferUtility ()
		{		
			if (FirstOfferGiven == false)
			{
				FirstOfferGiven = true;
				return 1;
			}
			
			double Min = Math.pow(GetDiscountFactor (), T.getTime());
			double Max = Min  * ( 1.5 + 5 * Noise) ;
			
			if ( T.getTime() < 0.88)
				Min *= Math.max(BestEverOpponentBidUtil, 0.75); 
			else if ( T.getTime() <= 0.94)
			{
				Min *= Math.max(BestEverOpponentBidUtil, 0.55); // never ever accept an offer with less than that utility
				Max *= 5; // slow down the decreasing
			}
			else if ( T.getTime() <= 0.975)
			{
				Min *= Math.max(BestEverOpponentBidUtil, 0.4); // never ever accept an offer with less than that utility
				Max *= 5; // slow down the decreasing
			}
			else
			{
				Max *= 20; // slow down the decreasing
				Min *= Math.max(2.5 * (1.0 / PlayerCount),MyUtilityOnOpponentBestBid); // never ever accept an offer with less than that utility
			}
			
			// the more eager the opponent is to settle, the slower we give up our utility. 
			// the higher the discount factor loss, the faster we give it up
			
			Max = Math.max(Max, Min);
			NextOfferUtility = Math.min(1, Max - (Max - Min) * T.getTime());
			return NextOfferUtility;
		}

		public double 	GetMinimumUtilityToAccept () // changes over time
		{
			return 0.9 * GetNextRecommendedOfferUtility ();
		}
		
		public void ProcessOpponentFirstBid(Bid firstBid)
		{
			
		}
		
		int CountdownToNoiseReestimation;
		String EstimateValue (Issue I, Value v) throws Exception
		{
			switch (I.getType())
			{
				case DISCRETE:
					ValueDiscrete DV = (ValueDiscrete)v;
					return DV.getValue();
				case INTEGER:
					int ValueIndex = 0;
					IssueInteger II = (IssueInteger)I;
					ValueInteger IV = (ValueInteger)v;
					double Step = II.getUpperBound() - II.getLowerBound();
					if (Step != 0)
					{
						int totalSteps= (int) Math.min(DiscretisationSteps, II.getUpperBound() - II.getLowerBound() + 1);
						Step /= totalSteps;
						ValueIndex = (int)(IV.getValue() / Step);
					}
					return String.valueOf(ValueIndex);
				case REAL:
					IssueReal RI = (IssueReal)I;
					ValueReal RV = (ValueReal)v;
					double StepR = RI.getUpperBound() - RI.getLowerBound();
					ValueIndex = 0;
					if (StepR != 0)
					{
						StepR /= DiscretisationSteps;
						ValueIndex = (int)(RV.getValue() / StepR);
					}
					return String.valueOf(ValueIndex);
			}
			
			throw new Exception("illegal issue");
		}
	
		
		TreeMap<Integer, TreeMap<String/*value.ToString()*/,Integer> > IncomingValueAppeareancesByIssueNum = new TreeMap<Integer, TreeMap<String,Integer>>(); // tells how many times each value has appeared in all incoming bids
		TreeMap<Integer, TreeMap<String/*value.ToString()*/,Integer> > OutgoingValueAppeareancesByIssueNum = new TreeMap<Integer, TreeMap<String,Integer>>(); // tells how many times each value has appeared in all incoming bids
		TreeMap<Integer, Integer > 						  DifferentValuesCountPerIssueNum = new TreeMap<Integer, Integer>(); // tells how many different values are per Issue
		int 											  CountdownToStatisticsRefreshing = 20; // when hits zero, recalculate AverageDistance and reset Noise
		double 											  PreviousAverageDistance = 0.2; // 0 to 1
		double 											  PreviousCountdownOpponentBestBidUtil = 1;
		double 											  BestEverOpponentBidUtil = 0;
		
		int 			OutgoingBidsCount = 0;		
		void 			AddMyBidToStatistics (Bid OutgoingBid) throws Exception
		{
			++OutgoingBidsCount;
			for (Issue I : US.getDomain().getIssues())
			{
				String bidValueEstimation = EstimateValue(I,OutgoingBid.getValue(I.getNumber()));
				if ( OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).containsKey(bidValueEstimation) )
				{
					OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).
						put(bidValueEstimation, 
								OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).get(bidValueEstimation) + 1);
				}
				else
				{
					OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).
						put(bidValueEstimation, 1); 
				}
			}
			
			
			
		}

		final int CountdownLength = 30;
		// update noise here
		public void 	ProcessOpponentBid (Bid IncomingBid) throws Exception
		{
			if (CountdownToStatisticsRefreshing > 0)
			{
				BestEverOpponentBidUtil = 
					Math.max (BestEverOpponentBidUtil, US.getUtility(IncomingBid));
					
				--CountdownToStatisticsRefreshing;
				for (Issue I : US.getDomain().getIssues())
				{
					String bidValueEstimation = EstimateValue(I,IncomingBid.getValue(I.getNumber()));
					if ( IncomingValueAppeareancesByIssueNum.get(I.getNumber()).containsKey(bidValueEstimation) )
					{
						IncomingValueAppeareancesByIssueNum.get(I.getNumber()).
							put(bidValueEstimation, 
								IncomingValueAppeareancesByIssueNum.get(I.getNumber()).get(bidValueEstimation) + 1);
					}
					else
					{
						IncomingValueAppeareancesByIssueNum.get(I.getNumber()).
							put(bidValueEstimation, 1);
					}
				}
			}
			else
			{
				double CurrentSimilarity = 0;
				
				for (Issue I : US.getDomain().getIssues())
				{
					for (String val : OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).keySet() )
					{
						if (IncomingValueAppeareancesByIssueNum.get(I.getNumber()).containsKey(val))
						{
							float outgoingVal = ((float)(OutgoingValueAppeareancesByIssueNum.get(I.getNumber()).get(val))) / OutgoingBidsCount;
							float incomingVal = (((float)(IncomingValueAppeareancesByIssueNum.get(I.getNumber()).get(val))) / CountdownLength);
							float diff = outgoingVal - incomingVal;
							float diffSqr = diff * diff; // 0 to 1
								
							CurrentSimilarity += 
								(1.0 / US.getDomain().getIssues().size() ) * 
								(1.0 / DifferentValuesCountPerIssueNum.get (I.getNumber()) ) * 
								(1 - diffSqr);
						}
					}
				}
				
				if (CurrentSimilarity > PreviousAverageDistance ||
					BestEverOpponentBidUtil > PreviousCountdownOpponentBestBidUtil)
					Noise += 0.02;
				else 
					Noise -= 0.05;

				Noise = Math.min ( Math.max(Noise, 0), 1);
				PreviousAverageDistance = CurrentSimilarity;
				CountdownToStatisticsRefreshing = CountdownLength;
				PreviousCountdownOpponentBestBidUtil = BestEverOpponentBidUtil;
				ClearIncommingStatistics();
			}
		}
		
		public Bid GenerateBidWithAtleastUtilityOf (double MinUtility)
		{
//			System.out.print("searching for :");
//			System.out.println (MinUtility);
			
			 Entry<Double,Bid> e = Bids.ceilingEntry(MinUtility);
			//System.out.print("found : " );
			
			 if (e == null)
			 {
				 //System.out.print(" not found :(  ");
				 try {
					//System.out.println(US.getUtility(this.maxBid));
				} catch (Exception e1) {
					// TODO Auto-generated catch block
					//e1.printStackTrace();
				}
				 return this.maxBid;
			 }
			
			 //System.out.println(e.getKey());
			return e.getValue();
		}

	}

	IssueManager IM;
	@Override
	public void init() 
	{
		// TODO Auto-generated method stub
		super.init();
		IM = new IssueManager(utilitySpace, timeline);
	}

	public String getName()
	{
		return "Gahboninho V2";
	}
	Bid previousBid = null;
	Bid OpponentBid = null;
	@Override
	public void ReceiveMessage(Action opponentAction) 
	{
		this.previousBid = this.OpponentBid;
		
		if (opponentAction instanceof Offer)
		{
			OpponentBid = ((Offer)opponentAction).getBid();
			if (this.previousBid != null)
			{
					try {
						this.IM.ProcessOpponentBid(this.OpponentBid);
					} catch (Exception e) {
						// Too bad
					}
			} 
			else
			{
				try {this.IM.learnBids(this.OpponentBid);} // learn from the first opp. bid
				catch (Exception e) {}
			}
		}
	}

	int FirstActions = 19;
	@Override
	public Action chooseAction() 
	{
		//System.out.println("\n\n");
		try 
		{
			if (OpponentBid != null && utilitySpace.getUtility(OpponentBid) >= IM.GetMinimumUtilityToAccept())
			{
				//System.out.print("I accept his offer of " + OpponentBid.toString() + "\nwith utility of :");
				//System.out.println(utilitySpace.getUtility(OpponentBid));
				return new Accept(this.getAgentID());
			}
		}
		catch (Exception e) 
		{
			// TODO Auto-generated catch block
			//e.printStackTrace();
		}
			
		// on the first time we act offer max bid
		if (this.previousBid == null)
		{
			try { IM.AddMyBidToStatistics (this.IM.getMaxBid()); } catch (Exception e2) {}
			return new Offer(this.getAgentID(),this.IM.getMaxBid());
		}
		
		Bid myBid;
		if (FirstActions >= 0 && timeline.getTime() < 0.15)
		{
			// on first few bids let the opponent learn some more about our preferences
			myBid = IM.GenerateBidWithAtleastUtilityOf(0.9 + 0.005 * FirstActions);
			--FirstActions;
		}
		else
		{
			myBid = IM.GenerateBidWithAtleastUtilityOf ( IM.GetNextRecommendedOfferUtility());
		}
		
		try { IM.AddMyBidToStatistics (myBid); } catch (Exception e2) {}
		
		//System.out.println("sending offer : " + myBid.toString());
		//System.out.print("with util of : ");
		//try {System.out.println(utilitySpace.getUtility(myBid));}
		//catch (Exception e) {e.printStackTrace();}
		
		return new Offer(this.getAgentID(), myBid );
	}

	@Override
	public AgentID getAgentID() {
		// TODO Auto-generated method stub
		return new AgentID("Gahboninho");
	}

}
