package negotiation.AgentX;

import java.util.ArrayList;
import java.util.Collections;

import negotiator.Agent;
import negotiator.Bid;
import negotiator.BidIterator;
import negotiator.Domain;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.EndNegotiation;
import negotiator.actions.Offer;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.ValueDiscrete;


public class AgentX extends Agent {
	private Action actionOfPartner = null;
	private OpponentModel oppModel;
	private ArrayList<Bid> possibleBids = new ArrayList<Bid>();
	private ArrayList<Double> possibleBidUtil = new ArrayList<Double>();
	private int indexOfLastBid; // /Index of our last bid in possibleBids array
	private double ourLowestBid = 2.0;
	private double theirLowestBid = 2.0;
	private int allbids = 0;
	double qTime;
	private int startingBidIssue = 0;
	private int kalaiIndex=-1;
	private double firstUtil=-1;

	/**
	 * init is called when a next session starts with the same opponent.
	 */
	public void init() {

		oppModel = new OpponentModel(utilitySpace);
		
	}

	public static String getVersion() {
		return "0.666";
	}

	@Override
	public String getName() {
		return "Agent X";
	}

	public void ReceiveMessage(Action opponentAction) {
		actionOfPartner = opponentAction;
	}

	/**
	 * Decides what is the next action of our agent.
	 * In the case of this being the first action of the agent it calculates all the possible bids
	 * and initializes the indexOfLastBid.
	 */
	public Action chooseAction() {
		if (allbids == 0) {
			allBids();
			allbids = 1;
			//System.out.println("End time " + timeline.getTime());
			indexOfLastBid = possibleBids.size()-1;
			while(!testIndexOfFirstIssue(possibleBids.get(0), startingBidIssue)){
				startingBidIssue++;
			}
			//System.out.println("\n\nThis is index od first issue on a bid: " + startingBidIssue);
		}

		Action action = null;
		try {
			// First bet is our best known bid
			if (actionOfPartner == null) {
				Bid bestBid = utilitySpace.getMaxUtilityBid();
				action = (new Offer(getAgentID(), bestBid));
				indexOfLastBid = possibleBids.size()-1;
			}
			if (actionOfPartner instanceof Offer) {
				Bid partnerBid = ((Offer) actionOfPartner).getBid();
				
				// get current time
				double time = timeline.getTime();
				
				//System.out.println("Processing opponent bid...");
				oppModel.processBid(partnerBid, timeline.getTime());
				//System.out.println("Opponent bid processed");

				double utilOpponentBid = calculateUtilityOpponent(partnerBid);
								
				if (utilOpponentBid < theirLowestBid){
					theirLowestBid = utilOpponentBid;
				}
				
				double own_Util=utilitySpace.getUtility(partnerBid);
				
				if (firstUtil<0){
					firstUtil = own_Util;
				}				
				
				action = chooseBidAction();
				
				if (acceptBid(partnerBid, time)) {
					return new Accept(getAgentID());
				}
				
				double discount = utilitySpace.getDiscountFactor();
				if(discount!=0)
				{
					if(firstUtil==own_Util && time>(0+discount))
					{
						return (new EndNegotiation(getAgentID()));
					}
				}
				
			}
		} catch (Exception e) {
			//System.out.println("Exception in ChooseAction:" + e.getMessage());
			action = new Accept(getAgentID()); // best guess if things go wrong.
		}
		return action;
	}

	/**
	 * Uses the current state of knowledge to choose the best possible for our agent bid.
	 * Takes the knowledge given by our opponent model, time discount factors and the time
	 * remaining to the end of our negotiation.
	 * @return the next action to perform, or accept in case of error
	 */
	int kalaiBid() {
		double time = timeline.getTime();
		int lower=0;
		int upper=possibleBids.size()-1;
		if(kalaiIndex>-1 && time>0.0)
		{
			lower=Math.max(0,kalaiIndex-Math.round(possibleBids.size()/8)-1);
			upper=Math.min(possibleBids.size()-1,kalaiIndex+Math.round(possibleBids.size()/8)+1);
		}
		double max=-1;
		int maxi=possibleBids.size() - 1;
		for (int i = upper; i >= lower; i--) {
			double UtilityOpp = calculateUtilityOpponent(possibleBids.get(i));
			double UtilityOwn = possibleBidUtil.get(i);
			double min=Math.min(UtilityOpp,UtilityOwn);
			if ( min>max )
			{
				max=min;
				maxi=i;
			} 
		}
		return maxi;
	}
	
	private Action chooseBidAction() {
		//System.out.println("Choosing a bid with opponent model");
		kalaiIndex=kalaiBid();
		double kalai = Math.min(possibleBidUtil.get(kalaiIndex),
				calculateUtilityOpponent(possibleBids.get(kalaiIndex)))-0.05;
		//double kalai = possibleBidUtil.get(kalaiIndex)-0.05;
		double util;
		double rv = 0;
		int firsttime=1;
		while(true)
		{
			int nextPareto = nextParetoBid(firsttime);
			if(firsttime==1) firsttime=0;
			Bid nextBid = possibleBids.get(nextPareto);
			double time = timeline.getTime(); 
			double discount = utilitySpace.getDiscountFactor();
			
			//qTime has a range of 0 to 4.6
			if(discount != 0){
				//double squaredTime = Math.pow(time, .25);
				//qTime = squaredTime*10/discount;
				//util=(.99 - qTime/42);
				util=(kalai+(1-Math.pow(time,4))*(0.99-kalai))*Math.pow(discount,time);
			}
			else{
				qTime = (Math.log((1/(1.1-time))*(1/(1.1-time)))+0.191);
				util=(kalai+(1-Math.pow(time,4))*(0.99-kalai));
			}
			try{
				rv = utilitySpace.getReservationValue();
			} catch(Exception e){
				//e.printStackTrace();
				rv = 0;
			}
			try {
				//if (utilitySpace.getUtility(nextBid) > (.99 - qTime/42)) {
				if(utilitySpace.getUtility(nextBid) < rv){
					return (new EndNegotiation(getAgentID()));
				}
				if (utilitySpace.getUtility(nextBid) >= util) {
					if( utilitySpace.getUtility(nextBid) < ourLowestBid){
						ourLowestBid = utilitySpace.getUtility(nextBid);
						//System.out.println("our lowest is " + ourLowestBid);
					}
					indexOfLastBid = nextPareto;
					//return (new Offer(getAgentID(), nextBid));
				} else {
					return (new Offer(getAgentID(),possibleBids.get(indexOfLastBid)));
				}
			} catch (Exception e) {
				//e.printStackTrace();
				return (new Accept(getAgentID()));
			}
		}
		
	}

	/**
	 * Just an auxiliar funtion to calculate the index where issues start on a bid
	 * because we found out that it depends on the domain.
	 * @return true when the received index is the proper index
	 */
	private boolean testIndexOfFirstIssue(Bid bid, int i){
		try{
			ValueDiscrete valueOfIssue = (ValueDiscrete) bid.getValue(i);
		}
		catch (Exception e) {
			return false;
		}
		return true;
	}
	
	/**
	 * Calculates the utility to our opponnent of the bid received as a parameter
	 * using the current knowledge given by our opponent model
	 * @param bid
	 * @return utility value for our opponnent
	 */
	private double calculateUtilityOpponent(Bid bid) {
		
		double utility = 0;

		// Taking into account that Opponent Issue list is in the same our of the domain
		int nrIssues = utilitySpace.getDomain().getIssues().size();
		
		for (int i = 0; i < nrIssues; i++) {
			try {
				//It was needed to use an auxiliar variable startingBidIssue, because we found out it varies from domain to domain
				ValueDiscrete valueOfIssue = (ValueDiscrete) bid.getValue(i+startingBidIssue);
				double w = oppModel.getIssueProcessor().getWeightByIssue( (IssueDiscrete) utilitySpace.getDomain().getIssue(i));
				double eval = oppModel.getEvaluationOfValue(i, valueOfIssue);

				utility += w * eval;

			} catch (Exception e) {
				//e.printStackTrace();
			}
		}
		
		return utility;
	}

	/**
	 * Takes into account our current opponent model and computes the closest 
	 * pareto optimal bid to our last bet
	 * NOTE: Since possibleBids is already ordered we only need to find the best
	 * 		 bid to the opponent  in our array of bids
	 * @return the index of the closest pareto optimal bet in possibleBids
	 * @return index of our last bid if there is no other pareto optimal 
	 */
	int nextParetoBid(int firsttime) {
		Bid lastBid = possibleBids.get(indexOfLastBid);
		int ini=(possibleBids.size() - 1);
		if(firsttime!=1) ini=indexOfLastBid;
		double d = calculateUtilityOpponent(lastBid);
		for (int i= ini; i >= 0; i--) {
			double newUtility = calculateUtilityOpponent(possibleBids.get(i));
			if (newUtility > d)
				return i;
		}
		return indexOfLastBid;
	}

	/**
	 * Calculates all the possible bids in the current domain
	 * and saves them, ordered by the utility for our agent, 
	 * in the possibleBids array
	 */
	void allBids() {

		Domain domain = utilitySpace.getDomain();
		BidIterator iter = new BidIterator(domain);
		//initiate a check for the first bid
		int teller = 0;
		Double temputil = 0.0;
		//iterate over all possible bids
		while (iter.hasNext()) {
			Bid tempBid = iter.next();
			try {
				temputil = utilitySpace.getUtility(tempBid);
			} catch (Exception e1) {
				e1.printStackTrace();
			}

			if (teller == 0) { // first bid doesn't need to be ordered
				possibleBids.add(tempBid);
				possibleBidUtil.add(temputil);
				teller = 1;
				//System.out.println("Start time: " + timeline.getTime());
			}
			//Locate the bids position in the array
			int pos = Collections.binarySearch(possibleBidUtil, temputil);
			if (teller == 1) {

				if (pos >= 0) {
					possibleBids.add(pos, tempBid);
					possibleBidUtil.add(pos, temputil);
				} else {
					possibleBids.add(-pos - 1, tempBid);
					possibleBidUtil.add(-pos - 1, temputil);
				}
			}
		}

	}
	
	/**
	 * Function that evaluates whether to accept a bid, based on the bid's utility and the time left for negotiation
	 */

	boolean acceptBid(Bid bid, double time){
		
		double rv = 0;
		double discount = utilitySpace.getDiscountFactor();
		//System.out.println(discount);
		try{
			rv = utilitySpace.getReservationValue();
		}
		catch (Exception e){
			//e.printStackTrace();
			rv = 0;
		}
		
		
		double util_a;
		//kalaiIndex=kalaiBid();
		double kalai = Math.min(possibleBidUtil.get(kalaiIndex),
				calculateUtilityOpponent(possibleBids.get(kalaiIndex)))-0.05;
		//double kalai = possibleBidUtil.get(kalaiIndex)-0.05;
		if(discount !=  0){
			//double squaredTime = Math.pow(time, .25);
			//qTime = squaredTime*10/discount;
			//util_a=(.99 - qTime/42);
			util_a=(kalai+(1-Math.pow(time,4))*(0.99-kalai))*Math.pow(discount,time);
		}
		else{
			//qTime has a range of 0 to 4.6
			qTime = (Math.log((1/(1.1-time))*(1/(1.1-time)))+0.191);
			util_a=(kalai+(1-Math.pow(time,4))*(0.99-kalai));
		}
		//range from util = 1 to util = .908
		//double alwaysAccept = 1 - qTime/50;
		
		double alwaysAccept =util_a;
		//reservation value, from .7 to .6
		double res = 0.7 - time/10;
		double util = 0;
		double dutil = 0;

		try {
			util = utilitySpace.getUtility(bid);	
		} 
		catch (Exception e1) {
			//e1.printStackTrace();
		}
		
		if(util<=rv){
			return false;
		}
		
		//Simplest form of acceptance, utility higher than a certain number
		if(util > alwaysAccept){
			//System.out.println("ACCEPTANCE 1, ALWAYS ACCEPT, qTime = " + qTime);
			return true;
		}
		//2nd form of acceptance, for higher utility than any of our offers(including our anticipated counter offer)
		if(util >= ourLowestBid){
			//System.out.println("ACCEPTANCE 2, OUR LOWEST BID");
			return true;
		}
		
		//3rd form, if the offer is in the top XX% of our utilities, has a safety value if there is a huge utility drop. Range is between 1% and  1.92%
		double index = possibleBids.size()/100 + (possibleBids.size()/500)*qTime;
		double minUtil = 1;
		
		try {
			minUtil = utilitySpace.getUtility((Bid) possibleBids.get(possibleBids.size() -1- (int) index));
			//System.out.println("minutil = " + minUtil);
			
		}
		catch (Exception e) {
			//e.printStackTrace();
		}
		
		if((util > minUtil)  && (util > res)){
			//System.out.println("ACCEPTANCE 3, POSITION IN ARRAY WITH MINUTIL = " + minUtil + " AND INDEX = " + index +" , qTime = " + qTime);
			return true;
		}
		
		return false;

	}
}
