/**
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 */
package negotiation.group4;

import java.util.ArrayList;

import negotiator.Bid;
import negotiator.issue.Issue;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.ValueDiscrete;

/**
 * This class makes a representation of a bid that is useable in the
 * opponent model which works with a matrix
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 * 
 */
public class BidIssue {
	public int[] issueIndices;
	
	public BidIssue(Bid bid, ArrayList<Issue> issues) {
		issueIndices = new int[issues.size()];
		try {
			// For each issue, see which index the bid value for that issue represents
			for (int i = 0; i < issues.size(); i++) {
				// TODO: Warning: start of workaround for bug where issue index is not issue index that should
				// be provided to getValue(int index) of Bid!
				int firstPossibleIndexOfValueInBid = -1;
				boolean succeeded = false;
				while (!succeeded) {
					firstPossibleIndexOfValueInBid++;
					try {
						bid.getValue(firstPossibleIndexOfValueInBid);
						succeeded = true;
					} catch (Exception e) {
						// Nothing to see here, Bid.getValue() not correctly implemented
					}
				}
				
				// Get the ValueDiscrete of the bid ...
				ValueDiscrete valueOfBid = (ValueDiscrete)bid.getValue(i+firstPossibleIndexOfValueInBid);
				// ... find out which index the bid represents
				issueIndices[i] = indexOfValueInIssue(valueOfBid, (IssueDiscrete)issues.get(i));
			}
		} catch (Exception e) {
			// If this ever happens, Genius/settings are to blame...
			e.printStackTrace();
		}
	}
	
	/**
	 * Find the index in the row of a certain issue from the domain
	 * @param value from a bid
	 * @param issue that should contain this value 
	 * @return index
	 */
	private int indexOfValueInIssue(ValueDiscrete value, IssueDiscrete issue) {
		int indexInIssue = -1;
		
		ArrayList<ValueDiscrete> values = issue.getValues();
		for (int i = 0; i < issue.getNumberOfValues() && indexInIssue == -1; i++) {
			if (values.get(i).value.equals(value.value)) {
				indexInIssue = i;
			}
		}
		
		return indexInIssue;
	}

}
