package travelingsalesman;

import java.util.ArrayList;
import negotiator.Bid;
import negotiator.Domain;
import negotiator.Timeline;
import negotiator.utility.UtilitySpace;
import negotiator.issue.Issue;

/**
 * This is an abstract class which manages all the negotiation Session pertinent information to a single agent
 * 
 @author Alex Dirkzwager
 */
public class NegotiationSession {
	
	protected OutcomeSpace outcomeSpace;
	protected BidHistory opponentBidHistory;
	protected BidHistory ownBidHistory;
	protected Domain domain;
	protected double discountFactor;
	protected UtilitySpace utilSpace;
	protected Timeline timeline;
	private BidDetails maxBidDetails;

	public NegotiationSession(UtilitySpace utilitySpace, Timeline time){
		utilSpace = utilitySpace;
		timeline = time;
		domain = utilSpace.getDomain();
		discountFactor = utilSpace.getDiscountFactor();
		opponentBidHistory = new BidHistory();
		ownBidHistory = new BidHistory();
		try {
			Bid maxBid = utilitySpace.getMaxUtilityBid();
			double maxBidUtil = utilitySpace.getUtility(maxBid);
			maxBidDetails = new BidDetails(maxBid, maxBidUtil, 0);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * gives a list of bids which the opponent has offered
	 * @return a list of BidDetails
	 */
	public BidHistory getOpponentBidHistory(){
		return opponentBidHistory;
	}
	
	public BidHistory getOwnBidHistory(){
		return ownBidHistory;
	}
	
	public double getDiscountFactor() {
		return discountFactor;
	}
	
	public ArrayList<Issue> getIssues(){
		return domain.getIssues();
	}
	
	public Timeline getTimeline(){
		return timeline;
	}
	
	/**
	 * gets the normalized time (t = [0,1])
	 * @return time normalized
	 */
	public double getTime() {
		return timeline.getTime();
	}
	
	public UtilitySpace getUtilitySpace(){
		return utilSpace;
	}
	
	public OutcomeSpace getOutcomeSpace(){
		return outcomeSpace;
	}
	
	public void setOutcomeSpace(OutcomeSpace space) {
		this.outcomeSpace = space;
	}
	
	/**
	 * gets the maximum utility possible from the domain
	 * @return double 
	 */
	public BidDetails getMaxBidinDomain() {
		return maxBidDetails;
	}
}